# =====================================
# generator=datazen
# version=3.1.2
# hash=48a9f16af16af88514e495c1a03fb18b
# =====================================
"""
rcmpy - Package's default entry-point.
"""

# built-in
import sys

# internal
from rcmpy.entry import main

if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""
Common path lookups used by the package.
"""

# built-in
from contextlib import contextmanager
from os import environ
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Iterator, Union

# third-party
from vcorelib.dict import limited

# internal
from rcmpy import PKG_NAME
from rcmpy.xdg import user_cache, user_config, user_state


def default_state_directory() -> Path:
    """Returns the default user-state directory used by the package."""
    return user_state(PKG_NAME)


def default_config_directory() -> Path:
    """Returns the default config directory used by the package."""
    return user_config(PKG_NAME, "default")


def default_cache_directory() -> Path:
    """Returns the default cache directory used by the package."""
    return user_cache(PKG_NAME)


@contextmanager
def override_environ(key: str, value: Union[Path, str]) -> Iterator[None]:
    """Override an environment variable as a managed context."""

    with limited(environ, key, str(value)):  # type: ignore
        yield


@contextmanager
def override_environ_tempdir(key: str) -> Iterator[Path]:
    """
    Set an environment variable to a temporary directory while it exists in
    the managed context.
    """

    with TemporaryDirectory() as tmpdir:
        with override_environ(key, tmpdir):
            yield Path(tmpdir)



"""
A module implementing an interface for package-managed files.
"""

# built-in
from contextlib import suppress
from dataclasses import dataclass
from pathlib import Path
from shutil import copyfile
import sys
from typing import Any, Dict, Set

# third-party
from vcorelib.logging import LoggerType
from vcorelib.paths import rel


def set_exec_flags(path: Path) -> None:
    """Set the executable bits, but respect the 'read' bits."""
    mode = path.stat().st_mode
    path.chmod(mode | ((mode & 0o444) >> 2))


@dataclass
class ManagedFile:
    """
    A data structure for managed files specified in the configuration data.
    """

    template: str
    extra_templates: Set[str]

    directory: Path
    name: str

    link: bool
    executable: bool

    condition: str

    platforms: Set[str]

    @property
    def output(self) -> Path:
        """Get the full output path."""
        return self.directory.joinpath(self.name)

    @property
    def present(self) -> bool:
        """Determine if this file is currently present in the file system."""
        return self.output.is_file()

    @property
    def platform(self) -> bool:
        """Determine if the platform is correct for handling this file."""
        return not self.platforms or sys.platform in self.platforms

    def evaluate(self, env: Dict[str, Any]) -> bool:
        """Determine if this file should be handled."""
        return self.platform and eval(  # pylint: disable=eval-used
            self.condition, None, env
        )

    def update_root(self, root: Path) -> None:
        """
        If the output directory is a relative path, update it to be an
        absolute one based on the provided root directory.
        """

        if not self.directory.is_absolute():
            self.directory = root.joinpath(self.directory)
            assert self.directory.is_absolute(), self.directory

    def update(self, source: Path, logger: LoggerType) -> None:
        """Update this managed file based on the provided source file."""

        output = self.output

        # At some point this could be skipped for existing symlinks.
        with suppress(FileNotFoundError):
            output.unlink()

        # Ensure the output directory exists.
        self.directory.mkdir(parents=True, exist_ok=True)

        if self.link:
            output.symlink_to(source)
        else:
            copyfile(source, output)

        if self.executable:
            set_exec_flags(output)

        logger.info("'%s' -> '%s'.", rel(source), rel(output))



# =====================================
# generator=datazen
# version=3.1.2
# hash=0cb0fd902b7216726825ed86f95176e5
# =====================================

"""
This package's command-line entry-point application.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser
from argparse import Namespace as _Namespace
from typing import Optional as _Optional

# third-party
from vcorelib.args import CommandFunction as _CommandFunction
from vcorelib.args import app_args as _app_args

# internal
from rcmpy.commands.all import commands

COMMAND: _Optional[_CommandFunction] = None


def entry(args: _Namespace) -> int:
    """Execute the requested task."""

    assert COMMAND is not None
    return COMMAND(args)


def add_app_args(parser: _ArgumentParser) -> None:
    """Add application-specific arguments to the command-line parser."""
    global COMMAND  # pylint: disable=global-statement
    add, COMMAND = _app_args(commands, {})
    add(parser)



"""
An environment extension for working with template files.
"""

# built-in
from copy import copy
from pathlib import Path
from typing import Dict, NamedTuple, Optional, Set, Tuple

# third-party
from datazen.templates import environment
from jinja2 import Environment, FileSystemLoader, Template
from vcorelib.paths.info import FileChangeEvent
from vcorelib.paths.info_cache import FileChanged, file_info_cache

# internal
from rcmpy.config import ManagedFile
from rcmpy.environment.base import BaseEnvironment
from rcmpy.environment.data import system_data


class EnvTemplate(NamedTuple):
    """
    A data structure for keeping track of environment templates. If 'template'
    is None, it signals that the template does not require rendering.
    """

    name: str
    path: Path
    subdir: str
    template: Optional[Template] = None


def template_name(name: str) -> Tuple[bool, str]:
    """
    Determine if a name is a template and remove the template suffix if
    applicable.
    """

    is_template = name.endswith(".j2")
    if is_template:
        name = name[:-3]
    return is_template, name


class TemplateEnvironment(BaseEnvironment):
    """
    A class implementing template management for this package's runtime
    environment.
    """

    templates: Dict[Path, EnvTemplate]
    templates_by_name: Dict[str, EnvTemplate]
    jinja: Environment

    def is_updated(self, file: ManagedFile) -> bool:
        """
        Determine if a managed file is updated or not (based on template
        changes).
        """
        return file.template in self.updated_template_names or any(
            x in self.updated_template_names for x in file.extra_templates
        )

    def _load_templates(self) -> None:
        """Load template information to the environment."""

        # Keep track of templates by name.
        self.templates = {}
        self.templates_by_name = {}
        for name in self.jinja.list_templates():
            obj = self.jinja.get_template(name)

            # We know all templates are files.
            assert obj.filename is not None

            is_template, name = template_name(name)

            # Keep track of template paths and their templates.
            path = Path(obj.filename).resolve()

            new = EnvTemplate(
                name,
                path,
                path.parent.name,
                obj if is_template else None,
            )
            self.templates[path] = new
            self.templates_by_name[name] = new

    def _init_templates(self, template_names: Set[str]) -> bool:
        """
        Initialize the template environment based on a set of template names
        that may be relevant to some task.
        """

        # The variant's template directory takes precedence.
        candidates = self.state.root_directories(
            "templates", common_first=False
        )

        # Prefer variant templates, if the variant template-directory
        # exists.
        self.jinja = environment(
            loader=FileSystemLoader(candidates, followlinks=True)
        )

        self._load_templates()

        def poll_cb(change: FileChanged) -> bool:
            """Aggregate paths of templates that have been updated."""

            if change.event is not FileChangeEvent.REMOVED:
                assert change.new is not None
                path = change.new.path

                subdir = path.parent.name
                _, name = template_name(path.name)

                # Don't acknowledge changes to any files that aren't active
                # templates.
                if (
                    name not in template_names
                    or name not in self.templates_by_name
                ):
                    return False

                template = self.templates_by_name[name]

                # If a version of a template that's not currently used has
                # changed, don't acknowledge the change either.
                if subdir != template.subdir:
                    return False

                self.updated_templates.add(path)
                self.updated_template_names.add(self.templates[path].name)

            return True

        template_changes = self.stack.enter_context(
            file_info_cache(
                self._cache.joinpath("templates.json"),
                poll_cb,
                logger=self.logger,
            )
        )

        # Poll template directories.
        for candidate in candidates:
            template_changes.poll_directory(candidate)

        # Log info about detected template changes.
        for changed in self.updated_templates:
            template = self.templates[changed]
            self.logger.info(
                "Detected change for template '%s' (%s).",
                template.name,
                template.subdir,
            )

        return True

    def _init_loaded(self) -> bool:
        """Called during initialization if a valid configuration is loaded."""

        result = super()._init_loaded()

        # Ensure double initialization isn't possible (use an arbitrary
        # attribute set here to detect this).
        assert not hasattr(self, "jinja")

        # Templates that are newly updated on this iteration.
        self.updated_templates: Set[Path] = set()
        self.updated_template_names: Set[str] = set()

        # Add additional information to template data.
        self.env_data = system_data()
        self.template_data = copy(self.state.configs)
        assert "env" not in self.template_data
        self.template_data["env"] = self.env_data

        return result and self._init_templates(self.config.templates)



"""
A module implementing a basic environment interface.
"""

# built-in
from contextlib import ExitStack
from pathlib import Path
from typing import Optional

# third-party
from vcorelib.io import ARBITER
from vcorelib.io.types import FileExtension, LoadResult
from vcorelib.logging import LoggerMixin, LoggerType
from vcorelib.paths import Pathlike, normalize, rel

# internal
from rcmpy import PKG_NAME
from rcmpy.config import Config
from rcmpy.paths import default_cache_directory
from rcmpy.state import State


def load_if_single_candidate(
    path: Pathlike, logger: LoggerType = None
) -> Optional[LoadResult]:
    """
    Attempt to load a configuration file if a candidate exists at the given
    path.
    """

    result = None
    path = normalize(path)

    config_candidates = list(
        FileExtension.data_candidates(path, exists_only=True)
    )

    if not config_candidates:
        if logger is not None:
            logger.error(
                "No config found: %s.",
                list(str(x) for x in FileExtension.data_candidates(path)),
            )
    elif len(config_candidates) == 1:
        result = ARBITER.decode(
            config_candidates[0],
            includes_key="includes",
            expect_overwrite=True,
        )
        if logger is not None:
            logger.info("Loaded config '%s'.", rel(config_candidates[0]))

    return result


def load_manifest(
    root: Path, variant: str, logger: LoggerType
) -> Optional[LoadResult]:
    """Load the top-level data repository configuration."""

    # Attempt to load the base manifest.
    config_data = load_if_single_candidate(
        root.joinpath(PKG_NAME), logger=logger
    )

    if config_data is not None:
        # Attempt to load variant-specific manifest data.
        variant_data = load_if_single_candidate(
            root.joinpath("includes", variant)
        )
        if variant_data is not None:
            config_data.merge(variant_data, expect_overwrite=True)

    return config_data


class BaseEnvironment(LoggerMixin):
    """A class implementing this package's base runtime environment."""

    def __init__(self, state: State, stack: ExitStack) -> None:
        """Initialize this instance."""

        super().__init__()
        self.state = state
        self.stack = stack
        self._config: Optional[Config] = None
        self._cache = default_cache_directory()

        self.build = state.directory.joinpath("build")
        self.build.mkdir(exist_ok=True)

        config_data = load_manifest(
            state.directory, state.variant, self.logger
        )
        if config_data is not None:
            self._config = Config(data=config_data.data)

            # Ensure that any relative paths called out are relative to the
            # root directory of the data repository.
            self._config.update_root(state.directory)

            # Treat manifest changes as criteria for updating outputs.
            self.state.update_manifest(self._config.data)

            # Consider the config not loaded if initialization fails.
            #
            # **Add this back in if initialization can actually fail.**
            #
            # if not self._init_loaded():
            #     self.logger.info("Initialization failed!")
            #     self._config = None
            assert self._init_loaded()

    def _init_loaded(self) -> bool:
        """Called during initialization if a valid configuration is loaded."""
        return True

    @property
    def config_loaded(self) -> bool:
        """Determine if this environment has loaded a config."""
        return self._config is not None

    @property
    def config(self) -> Config:
        """Get this environment's configuration object."""
        assert self._config is not None, "Config not loaded!"
        return self._config



"""
A module implementing this package's stateful information that persists across
invocations.
"""

# built-in
from contextlib import ExitStack, contextmanager
from io import StringIO
from logging import getLogger
from pathlib import Path
from typing import Any, Dict, Iterator, List, cast

# third-party
from jinja2 import Template
from vcorelib.dict import merge
from vcorelib.dict.cache import FileCache
from vcorelib.io import ARBITER
from vcorelib.io.types import DataStream
from vcorelib.io.types import JsonObject as _JsonObject
from vcorelib.paths import Pathlike, normalize, rel

# internal
from rcmpy.paths import default_config_directory, default_state_directory
from rcmpy.schemas import RcmpyDictCodec as _RcmpyDictCodec

LOG = getLogger(__name__)


class State(_RcmpyDictCodec):
    """The top-level configuration object for the package."""

    directory: Path
    variant: str
    variables_new: bool
    configs_new: bool
    manifest: Dict[str, Any]
    manifest_new: bool

    def init(self, data: _JsonObject) -> None:
        """Perform implementation-specific initialization."""

        self.logger = LOG

        self.directory = normalize(
            cast(str, data.get("directory", default_config_directory()))
        ).resolve()

        self.logger.info("Using directory '%s'.", rel(self.directory))

        self.variant: str = cast(str, data["variant"])
        if self.variant:
            self.logger.info("Using variant '%s'.", self.variant)

        data.setdefault("previous", {})
        self.previous: Dict[str, Any] = data["previous"]  # type: ignore
        self.previous.setdefault("variant", "default")

        # Variables.
        self.variables: Dict[str, Any] = {}
        self.variables_new = False
        self.previous.setdefault("variables", {})
        self._load_variables()

        # Configs.
        self.configs: Dict[str, Any] = {}
        self.configs_new = False
        self.previous.setdefault("configs", {})
        self._load_configs()

        # Manifest configuration.
        self.manifest: Dict[str, Any] = cast(
            Dict[str, Any], data.get("manifest", {})
        )
        self.manifest_new: bool = False

    def is_new(self) -> bool:
        """Determine if state has changed."""
        return (
            self.variant != self.previous["variant"]
            or self.variables_new
            or self.configs_new
            or self.manifest_new
        )

    def update_manifest(self, data: Dict[str, Any]) -> None:
        """Set new manifest data."""
        self.manifest_new = self.manifest != data
        self.manifest = data

    def root_directories(
        self, subdir: str, common_first: bool = True
    ) -> List[Path]:
        """
        Get up to a pair of directories from some sub-directory of the current
        root.
        """

        root = self.directory.joinpath(subdir)

        candidates = [root.joinpath("common")]
        if self.variant:
            if common_first:
                candidates.append(root.joinpath(self.variant))
            else:
                candidates.insert(0, root.joinpath(self.variant))

        return [x for x in candidates if x.is_dir()]

    def _load_configs(self) -> None:
        """Load data for configs."""

        with ExitStack() as stack:

            def preprocessor(stream: DataStream) -> DataStream:
                """Render the variable file as a template."""

                return stack.enter_context(
                    StringIO(Template(stream.read()).render(self.variables))
                )

            for path in self.root_directories("configs"):
                merge(
                    self.configs,
                    ARBITER.decode_directory(
                        path,
                        logger=self.logger,
                        require_success=True,
                        recurse=True,
                        includes_key="includes",
                        expect_overwrite=True,
                        preprocessor=preprocessor,
                    ).data,
                    expect_overwrite=True,
                    logger=self.logger,
                )

        self.configs_new = self.configs != self.previous["configs"]
        self.previous["configs"] = self.configs

        if self.configs_new:
            self.logger.info("Configuration data is updated.")

    def _load_variables(self) -> None:
        """Load data for variables."""

        for path in self.root_directories("variables"):
            merge(
                self.variables,
                ARBITER.decode_directory(
                    path,
                    logger=self.logger,
                    require_success=True,
                    recurse=True,
                    includes_key="includes",
                    expect_overwrite=True,
                ).data,
                expect_overwrite=True,
                logger=self.logger,
            )

        self.variables_new = self.variables != self.previous["variables"]
        self.previous["variables"] = self.variables

        if self.variables_new:
            self.logger.info("Variable data is updated.")

    def set_directory(self, path: Pathlike) -> None:
        """Set a new directory to use as the data repository."""

        new_dir = normalize(path).resolve()

        if new_dir == self.directory:
            self.logger.info(
                "New directory '%s' same as current.", rel(new_dir)
            )
            return

        self.directory = new_dir
        self.logger.info("Set directory to '%s'.", rel(new_dir))

    def set_variant(self, variant: str = None) -> None:
        """Set a new variant value."""

        if not variant:
            self.logger.info("Current variant: '%s'.", self.variant)
        elif variant != self.variant:
            self.previous["variant"] = self.variant
            self.variant = variant
            self.logger.info("Updating variant to '%s'.", variant)

    def asdict(self) -> _JsonObject:
        """Obtain a dictionary representing this instance."""

        return {
            "directory": str(self.directory),
            "variant": self.variant,
            "previous": self.previous,
            "manifest": self.manifest,
        }


@contextmanager
def load_state(
    root: Pathlike = None, name: str = "state.json"
) -> Iterator[State]:
    """This needs to be a context manager, so it get's written back to disk."""

    if root is None:
        root = default_state_directory()

    with FileCache(normalize(root, name)).loaded() as data:
        state = State.create(data, verify=False)
        yield state

        # Update the original dictionary with any changes to the state object.
        data.update(state.asdict())



"""
A module for common argument-parsing methods.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser
from argparse import Namespace as _Namespace
from logging import getLogger
from typing import Callable

# third-party
from vcorelib.logging import log_time

# internal
from rcmpy.environment import Environment, load_environment


def add_default_flag(parser: _ArgumentParser) -> None:
    """Adds a default-flag argument."""

    parser.add_argument(
        "-d",
        "--default",
        action="store_true",
        help="sets the directory back to the package default",
    )


EnvCommand = Callable[[_Namespace, Environment], int]


def run_env_command(args: _Namespace, cmd: EnvCommand) -> int:
    """Run a command with the environment instance loaded."""

    result = 1

    with log_time(getLogger(__name__), "Command"):
        with load_environment() as env:
            if env.config_loaded:
                result = cmd(args, env)

    return result



"""
An entry-point for the 'use' command.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser
from argparse import Namespace as _Namespace
from pathlib import Path

# third-party
from vcorelib.args import CommandFunction as _CommandFunction

# internal
from rcmpy.commands.common import add_default_flag
from rcmpy.paths import default_config_directory
from rcmpy.state import load_state


def use_cmd(args: _Namespace) -> int:
    """Execute the use command."""

    with load_state() as state:
        if args.default:
            args.directory = default_config_directory()

        if args.directory is not None:
            state.set_directory(args.directory)

    return 0


def add_use_cmd(parser: _ArgumentParser) -> _CommandFunction:
    """Add use-command arguments to its parser."""

    add_default_flag(parser)

    parser.add_argument(
        "directory", type=Path, nargs="?", help="the directory to use"
    )

    return use_cmd



"""
An entry-point for the 'dump' command.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser
from argparse import Namespace as _Namespace
from json import dump
import sys

# third-party
from vcorelib.args import CommandFunction as _CommandFunction

# internal
from rcmpy.commands.common import run_env_command
from rcmpy.environment import Environment


def dump_env(_: _Namespace, env: Environment) -> int:
    """Dump information about the environment."""

    dump(env.template_data, sys.stdout, indent=2, sort_keys=True)
    return 0


def dump_cmd(args: _Namespace) -> int:
    """Execute the dump command."""
    return run_env_command(args, dump_env)


def add_dump_cmd(_: _ArgumentParser) -> _CommandFunction:
    """Add dump-command arguments to its parser."""

    return dump_cmd



"""
rcmpy - Test the program's entry-point.
"""

# built-in
from subprocess import check_output
from sys import executable
from unittest.mock import patch

# module under test
from rcmpy import PKG_NAME
from rcmpy.entry import main as rcmpy_main


def test_entry_basic():
    """Test basic argument parsing."""

    args = [PKG_NAME]
    assert rcmpy_main(args + ["-h"]) == 0

    with patch("rcmpy.entry.entry", side_effect=SystemExit(1)):
        assert rcmpy_main(args) != 0


def test_package_entry():
    """Test the command-line entry through the 'python -m' invocation."""

    check_output([executable, "-m", "rcmpy", "-h"])



"""
A module implementing a configuration interface for the package.
"""

# built-in
from os.path import expandvars
from pathlib import Path
from typing import Any, Dict, List, Set, cast

# third-party
from vcorelib.dict.codec import BasicDictCodec as _BasicDictCodec
from vcorelib.io.types import JsonObject as _JsonObject

# internal
from rcmpy.config.file import ManagedFile
from rcmpy.schemas import RcmpyDictCodec as _RcmpyDictCodec

FilesConfig = List[Dict[str, Any]]

__all__ = ["FilesConfig", "Config", "ManagedFile"]


class Config(_RcmpyDictCodec, _BasicDictCodec):
    """The top-level configuration object for the package."""

    def update_root(self, root: Path) -> None:
        """Call 'update_root' for each managed file."""

        for file in self.files:
            file.update_root(root)

    def init(self, data: _JsonObject) -> None:
        """Initialize this instance."""
        self.data = data

        # A set of templates that are used by this config.
        self.templates: Set[str] = set()

        self.files: List[ManagedFile] = []
        for file in cast(FilesConfig, data.get("files", [])):
            new = ManagedFile(
                file["template"],
                set(file.get("extra_templates", [])),
                # Resolve environment variables and any '~' here.
                Path(expandvars(file["directory"])).expanduser(),
                file.get("name", file["template"]),
                file["link"],
                file["executable"],
                file["condition"],
                set(file.get("platforms", [])),
            )

            # Keep track of all templates used in any file.
            self.templates.add(new.template)
            for template in new.extra_templates:
                self.templates.add(template)

            self.files.append(new)



"""
A module for exposing useful information from the current system to templates.
"""

# built-in
from functools import lru_cache
import platform
import sys
from typing import Any, Dict


@lru_cache(maxsize=1)
def system_data() -> Dict[str, Any]:
    """Obtain useful system configuration data."""

    return {
        "sys": {
            "base_exec_prefix": sys.base_exec_prefix,
            "base_prefix": sys.base_prefix,
            "byteorder": sys.byteorder,
            "exec_prefix": sys.exec_prefix,
            "executable": sys.executable,
            "platform": sys.platform,
            "prefix": sys.prefix,
            "version": sys.version,
        },
        "platform": {
            "bits": platform.architecture()[0],
            "linkage": platform.architecture()[1],
            "machine": platform.machine(),
            "node": platform.node(),
            "processor": platform.processor(),
            "implementation": platform.python_implementation(),
            "python_version": platform.python_version(),
            "release": platform.release(),
            "system": platform.system(),
            "version": platform.version(),
        },
    }



"""
An entry-point for the 'apply' command.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser
from argparse import Namespace as _Namespace

# third-party
from vcorelib.args import CommandFunction as _CommandFunction
from vcorelib.paths import rel

# internal
from rcmpy.commands.common import run_env_command
from rcmpy.environment import Environment


def apply_env(args: _Namespace, env: Environment) -> int:
    """Apply pending changes from the environment."""

    result = 0

    is_new = env.state.is_new()

    for file in env.config.files:
        # Check if a template is found for this file.
        if file.template not in env.templates_by_name:
            result += 1
            env.logger.error("Template '%s' not found!", file.template)
            continue

        if not file.evaluate(env.env_data):
            continue

        # Check if this file has any updated templates.
        if args.force or is_new or not file.present or env.is_updated(file):
            template = env.templates_by_name[file.template]

            # If a template doesn't require rendering, use it as-is.
            source = template.path

            if template.template is not None:
                # Render the template to the build directory.
                source = env.build.joinpath(file.template)
                with source.open("w") as path_fd:
                    path_fd.write(template.template.render(env.template_data))
                    env.logger.info("Rendered '%s'.", rel(source))

            # Update the output file.
            if not args.dry_run:
                file.update(source, env.logger)

    return result


def apply_cmd(args: _Namespace) -> int:
    """Execute the apply command."""
    return run_env_command(args, apply_env)


def add_apply_cmd(parser: _ArgumentParser) -> _CommandFunction:
    """Add apply-command arguments to its parser."""

    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="whether or not to forcibly render all outputs",
    )
    parser.add_argument(
        "-d",
        "--dry-run",
        action="store_true",
        help="whether or not to update output files",
    )

    return apply_cmd



# =====================================
# generator=datazen
# version=3.1.2
# hash=c3c3f1e4bd14e6c9f1861b5a32713a39
# =====================================

"""
Useful defaults and other package metadata.
"""

DESCRIPTION = "A configuration-file management system."
PKG_NAME = "rcmpy"
VERSION = "1.4.1"
